"Wrong code."

# from pandas import DataFrame
# from sklearn.model_selection import train_test_split
#
# from linear_regression import LinearRegressor
# """
# Logistic function == sigmoid function
# Some assumptions of logistic regression :
#     1) All predictors are independent of each other.
#     2) Each predictor must atleast have 50 rows each for better output.
# """
#
# """
#     S curve is linear in logit
#     Logit  = log of odds"
#     Logit(p) = log(p/(1-p))
# """
#
# class Model:
#     def __init__(self, regressor, reg_coeff):
#         self.regressor = regressor
#         self.b1, self.b0 = reg_coeff
#         b1, b0 = self.b1, self.b0
#         print("b1, b0 : ", b1, b0)
#         e_b1_b0x = lambda x : __import__("math").exp(b1*x + b0)
#
#         self.coeff = lambda x : e_b1_b0x(x)/\
#                                 (e_b1_b0x(x)+1)
#     def predict(self, test_values):
#         predictions = []
#         for x in test_values :
#             predictions.append(self.coeff(x))
#         return predictions
#
# class LogisticRegressor:
#     """
#         Data will have last column as target variable.
#         For simplicity, considering only one feature.
#     """
#
#     def __init__(self, data_set, test_size=0.2):
#         self.data_set = DataFrame(
#             {
#                 'x': data_set[0],  # represents x axes points
#                 'y': data_set[1]
#             }
#         )
#         self.train, self.test = train_test_split(self.data_set, test_size=test_size)
#
#     def fit(self):
#         data_set = [list(self.train["x"]), list(self.train["y"])]
#         linear_regressor = LinearRegressor(data_set)
#         linear_model = linear_regressor.fit()
#         b1, b0 = linear_model.reg_coeff        #b is beta.
#         e = __import__("math").e
#
#         #logit(p) = b1*x+b0
#         self.model = Model(self, [b1, b0])
#         return self.model
#
# data_set = [list(range(10)), [1 if i%5 == 0 else 0 for i in range(10)]]
# print(data_set[1])
# regressor = LogisticRegressor(data_set)
# model = regressor.fit()
# print(model.predict([5, 10, 12]))


from sklearn.model_selection import train_test_split
from numpy import array


class LinearRegressor:
    def __init__(self, dataset, error_limit=10**(-10)):
        """
        Assumptions :
            1) last row is target
        """
        self.dataset = dataset
        self.error_limit = error_limit
        self.train_data, self.test_data = train_test_split(dataset, shuffle=False)

        self.x = [[1] + row[:-1] for row in self.train_data]  # [1] is for  bias.
        self.y = [row[-1] for row in self.train_data]

        self.test_x = [[1] + row[:-1] for row in self.test_data]  # [1] is for  bias.
        self.test_y = [row[-1] for row in self.test_data]

        self.n = len(self.x[0])  # n is # features.
        self.m = len(self.x)

        self.θ = array([0 for i in range(self.n)])

    def hθ(self, x):
        return self.θ.dot(x)

    def Jθ(self):  # input should be regression coefficient
        return (1 / (2 * self.m)) * \
               sum(
                   [(self.hθ(self.x[i]) - self.y[i]) ** 2
                    for i in range(self.m)]
               )

    def Jθ_derivative(self, j):
        return (1 / (self.m)) * \
               sum(
                   [
                       (self.hθ(self.x[i]) - self.y[i]) * self.x[i][j]
                       for i in range(self.m)
                   ]
               )

    def is_convergent(self):
        cost = self.Jθ()
        if cost >= 0:
            if cost < self.error_limit:
                return True

    def fit(self, α=0.1, max_epochs=1000):
        # α : learning rate
        epoch_number = 0
        while True:
            if self.is_convergent(): break
            if epoch_number and epoch_number % max_epochs == 0:
                print(epoch_number, "epochs elapsed")
                print(self.accuracy(), "is the current accuracy.")
                stop = 1 if input("Do you want to stop training (y/*)??").lower() == 'y' else 0
                if stop:
                    print("Training aborted.")
                    break
            temp = []
            for j in range(len(self.θ)):
                temp.append(self.θ[j] - α * self.Jθ_derivative(j))
            self.θ = array(temp)
            epoch_number += 1
        print("Training completed successfully. Required {} epochs".format(epoch_number))

    def predict(self, x=None):
        error = False
        if x is None:
            x = self.test_x
        try:
            x[0][0]
        except:
            "Expected 2d array."
            error = True
        try:
            if len(x[0]) != len(self.test_x[0]) - 1:
                "Incorrect number of feature variables."
                error = True
        except:
            error = True
        if error:
            "Error occured, Predicting with test data."
            x = self.test_x
        result = []
        for x in x:
            if not error:
                x = [1]+x  # [1] is for  bias.
            result.append(self.hθ(x))
        return result


    def accuracy(self, predictions=None, expectations=None):

        if not predictions or not expectations:
            # user wants to get accuracy based on test data
            train_data = self.test_x
            predictions = self.predict(self.test_x)
            expectations = list(self.test_y)

        n = len(min([predictions, expectations], key=len))   # if predictions and expectations are different in length
        p, e = predictions, expectations
        return 1-sum(
            [
                abs(p[i]-e[i])/e[i]
                for i in range(n)
                if e[i] != 0]
        )/n


if __name__ == '__main__':
    l = LinearRegressor([
        [2, 1, 2],                # table of two
        [2, 2, 4],
        [2, 3, 6],
        [2, 4, 8],
        [2, 5, 10]
    ])
    l.fit(max_epochs=200)

    test_data = [[2, 11], [2, 12], [2, 13], [2, 14]]
    target_data = [22, 24, 26, 28]
    predictions = l.predict(test_data)
    print(l.accuracy(predictions, target_data), "is the accuracy")


# Linear Regression for any number of parameters in dataset.

"""
    Assumptions :
        1) Each parameter is independent of each other.
        2) The last variable is target dataset[-1].
        3) No none/null/empty values in dataset
        4) No outliers in dataframe
"""

from seaborn import pairplot
from pandas import DataFrame
from matplotlib import pyplot as plt
from sklearn.model_selection import train_test_split
from numpy import multiply as mul_matrix
# from sklearn.metrics import confusion_matrix


class Model:
    def __init__(self, l_regressor, reg_coeff):
        self.regressor = l_regressor
        self.reg_coeff = reg_coeff

    def predict(self, test_values):
        """
            Expects : list of x
            Returns : list of predictions
            one extra parameter in reg_coeff list of linear shift
        """
        predictions = []
        for ele in test_values:
            predictions.append(mul_matrix(self.reg_coeff, [ele, 1]).sum())

        return predictions

    def accuracy(self, predictions=None, expectations=None):

        if not predictions or not expectations:
            # user wants to get accuracy based on test data
            train_data = self.regressor.train
            predictions = self.predict(train_data["x"])
            expectations = list(train_data["y"])

        n = len(min([predictions, expectations], key=len))   # if predictions and expectations are different in length
        p, e = predictions, expectations
        return 1-sum(
            [
                abs(p[i]-e[i])/e[i]
                for i in range(n)
                if e[i] != 0]
        )/n


class LinearRegressor:
    def __init__(self, data_set, test_size=0.2):
        self.data_set = DataFrame(
            {
                'x': data_set[0],  # represents x axes points
                'y': data_set[1]
            }
        )
        self.train, self.test = train_test_split(self.data_set, test_size=test_size, shuffle=False)

    def plot(self, data=None):
        if data is None:
            "default: plotting the given dataframe"
            pairplot(self.data_set)
        else:
            pairplot(data)
        plt.show()


    def fit(self):
        # Y = mX + c

        # getting variables ready :
        X = self.train['x']    # capital indicates : train data(larger one)
        Y = self.train['y']
        n = len(X)
        sum_X = X.sum()
        sum_Y = Y.sum()
        sum_XX = mul_matrix(X, X).sum()
        sum_YY = mul_matrix(Y, Y).sum()
        sum_XY = mul_matrix(X, Y).sum()

        m = (n*sum_XY - sum_X*sum_Y)/(n*sum_XX - sum_X**2)
        c = (sum_Y - m*sum_X)/n

        self.model = Model(self, [m, c])    # model has reg_coeff in decreasing order of power.
        return self.model


if __name__ == "__main__":
    x = [0, 1, 3, 3, 4, 5, 6, 7, 8, 9]          # line y=x with some jagginess.
    y = list(range(10))
    dataset = [x, y]
    regressor = LinearRegressor(dataset)

    model = regressor.fit()

    test_data = range(10)

    prediction = model.predict(test_data)

    print("Accuracy of model : ", model.accuracy(predictions=prediction, expectations=test_data))
